//
//
// #################################  Interface routines for
// #####  C Library Interface  #####  some C library routines.
// #################################
// 
// This interface tries to be compatible with different versions
// of libc (libc5 and glibc AKA libc6) and different versions of
// the Linux kernel.
//
// Functions xb_stat(), xb_sigaction(), xb_readdr() in this file
// provide the same function-interface and behavior as functions
// stat(), sigaction(), readdr() in libc5.  The function-interface
// and/or behavior of these functions changed from libc5 to libc6,
// which requires the following solution or changes to XBasic code.
//
// This file was created by Eddie Penninkhof to make Linux XBasic
// work on his Linux system and other releases of Linux based on
// the libc6 function libraries.  Thanks Eddie!
//
// PROGRAM "xbiface"
// VERSION "0.0002"
//
//
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <dirent.h>
//
// this is the stat struct recognized by XBasic
//
struct xb_newustat {
	unsigned short     st_dev;
	unsigned short     st_pad1;
	unsigned long      st_ino;
	unsigned short     st_mode;
	unsigned short     st_nlink;
	unsigned short     st_uid;
	unsigned short     st_gid;
	unsigned short     st_rdev;
	unsigned short     st_pad;
	unsigned long      st_size;
	unsigned long      st_blksize;
	unsigned long      st_blocks;
	unsigned long      st_atime;
	unsigned long      st_unused1;
	unsigned long      st_mtime;
	unsigned long      st_unused2;
	unsigned long      st_ctime;
	unsigned long      st_unused3;
	unsigned long      st_unused4;
	unsigned long      st_unused5;
};
//
// this is the stat struct previously recognized by XBasic
//
struct xb_oldustat {
	short							st_dev;
	unsigned short		st_ino;
	unsigned short		st_mode;
	short							st_nlink;
	unsigned short		st_uid;
	unsigned short		st_gid;
	short							st_rdev;
	unsigned short		st_pad;
	long							st_size;
	unsigned long			st_atime;
	unsigned long			st_mtime;
	unsigned long			st_ctime;
};
//
// this is the sigaction struct recognized by XBasic
//
struct xb_sigaction {
	long	xb_sa_handler;
	long	xb_sa_mask;
	long	xb_sa_flags;
};
//
// this is the dirent struct recognized by XBasic
//
struct xb_dirent {
	long		d_ino;
	long		d_off;
	unsigned short	d_reclen;
	unsigned short	d_pad;
	char		d_name[256];
};
//
//
// #######################
// #####  xb_stat()  #####
// #######################
//
int xb_stat(char *filename, struct xb_oldustat *xb_stat_buf) {
	int ret = 0;
	struct stat stat_buf;

	ret = stat(filename, &stat_buf);

	if (ret == 0) {
		xb_stat_buf->st_dev = stat_buf.st_dev;
//	xb_stat_buf->st_pad1 = 0
		xb_stat_buf->st_ino = stat_buf.st_ino;
		xb_stat_buf->st_mode = stat_buf.st_mode;
		xb_stat_buf->st_nlink = stat_buf.st_nlink;
		xb_stat_buf->st_uid = stat_buf.st_uid;
		xb_stat_buf->st_gid = stat_buf.st_gid;
		xb_stat_buf->st_rdev = stat_buf.st_rdev;
		xb_stat_buf->st_pad = 0;
		xb_stat_buf->st_size = stat_buf.st_size;
//	xb_stat_buf->st_blksize = stat_buf.st_blksize;
//	xb_stat_buf->st_blocks = stat_buf.st_blocks;
		xb_stat_buf->st_atime = stat_buf.st_atime;
//	xb_stat_buf->st_unused1 = 0;
		xb_stat_buf->st_mtime = stat_buf.st_mtime;
//	xb_stat_buf->st_unused2 = 0;
		xb_stat_buf->st_ctime = stat_buf.st_ctime;
//	xb_stat_buf->st_unused3 = 0;
//	xb_stat_buf->st_unused4 = 0;
//	xb_stat_buf->st_unused5 = 0;
	} else {
		memset(xb_stat_buf, 0, sizeof(*xb_stat_buf));
	}
	return (ret);
}
//
//
// ############################
// #####  xb_sigaction()  #####
// ############################
//
// The number of possible signals is extended from 32 to 1024,
// so sa_mask does not fit any longer in a long variable.
// This interface function converts the long into a sigset_t.
//
int xb_sigaction(int signum, const struct xb_sigaction *act, struct xb_sigaction *oldact) {
	struct sigaction oldact1;
	struct sigaction act1;
	long sigmask;
	int ret;
	int i;

	act1.sa_handler = (void (*)(int))act->xb_sa_handler;
	sigemptyset(&act1.sa_mask);
	sigmask = 1;
//
// we assume signal-numbering has not changed.
// somebody needs to make sure this is/stays true.
//
	for (i = 0; i < 32; ++i) {
		if ((act->xb_sa_mask & sigmask) != 0) { sigaddset(&act1.sa_mask, i + 1); }
		sigmask <<= 1;
	}

	act1.sa_flags = act->xb_sa_flags;
	act1.sa_restorer = NULL;
	ret = sigaction(signum, &act1, &oldact1);

	if (oldact != NULL) {
		oldact->xb_sa_handler = (long)oldact1.sa_handler;
		oldact->xb_sa_mask = 0;
		sigmask = 1;

		for (i = 0; i < 32; ++i) {
			if (sigismember(&oldact1.sa_mask, i + 1)) { oldact->xb_sa_mask |= sigmask; }
			sigmask <<= 1;
		}

		oldact->xb_sa_flags = oldact1.sa_flags;
	}
	return (ret);
}
//
//
// #########################
// #####  xb_readdr()  #####
// #########################
//
int xb_readdir(DIR *dirp, struct xb_dirent *dirent) {
	struct dirent	*dirent1;

	dirent1 = readdir(dirp);
	if (dirent1 == NULL) { return (0); }

	dirent->d_ino = dirent1->d_ino;
	dirent->d_off = dirent1->d_off;
	dirent->d_reclen = dirent1->d_reclen;
	dirent->d_pad = 0;

	memcpy(dirent->d_name, dirent1->d_name, 256);
	return (1);
}
